﻿:NameSpace Utils
⍝ Namespace containing :Includable utility functions

    ⎕IO ⎕ML←1  ⋄ Version←1.01

⍝ --- Selections ---

    where←{⍵/⍳⍴⍵}        ⍝ Indices of 1's in ⎕IO 1

⍝ --- Strings ---

    lc←'abcdefghijklmnopqrstuvwxyzàáâãåèéêëòóôõöøùúûäæüìíîïðçñ' ⍝ (lower case alphabet)
    uc←'ABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÅÈÉÊËÒÓÔÕÖØÙÚÛÄÆÜÌÍÎÏÐÇÑ' ⍝ (upper case alphabet)
    fromto←{n←⍴1⊃(t f)←⍺ ⋄ ~∨/b←n≥i←f⍳s←,⍵:s ⋄ (b/s)←t[b/i] ⋄ (⍴⍵)⍴s} ⍝ from-to casing fn

    lcase←lc uc∘fromto ⍝ :Includable Lower-casification of simple array
    ucase←uc lc∘fromto ⍝ Ditto Upper-casification

    cut←{⍺←⍵∊1↑⍵ ⋄ ⎕ml←0 ⋄ 1↓¨⍺⊂⍵}                  ⍝ exclude delimiters cut

      dtb←{                                         ⍝ Drop Trailing Blanks.
          ⍺←' ' ⋄ 1<|≡⍵:(⊂⍺)∇¨⍵                     ⍝ nested?
          2<⍴⍴⍵:(¯1↓⍴⍵){(⍺,1↓⍴⍵)⍴⍵}⍺ ∇,[¯1↓⍳⍴⍴⍵]⍵   ⍝ array
          1≥⍴⍴⍵:(-+/∧\⌽⍵∊⍺)↓⍵                       ⍝ vector
          (~⌽∧\⌽∧⌿⍵∊⍺)/⍵                            ⍝ matrix
      }

      dmb←{                                         ⍝ Drop Multiple Blanks.
          ⍺←' ' ⋄ 1<|≡⍵:(⊂⍺)∇¨⍵                     ⍝ nested?
          2<⍴⍴⍵:(¯1↓⍴⍵){(⍺,1↓⍴⍵)⍴⍵}⍺ ∇,[¯1↓⍳⍴⍴⍵]⍵   ⍝ array
          2>⍴⍴⍵:(2∨/(~⍵∊⍺),1)/⍵                     ⍝ vector
          (2∨/(,∨⌿~⍵∊⍺),1)/⍵                        ⍝ matrix
      }

      drvSrc←{⎕IO ⎕ML←0                       ⍝ JS 2012 Guess source of derived function.
     
          trav←{                                  ⍝ traverse, accumulating subtrees.
              ~(⊂⍴⍵)∊,¨2 3:leaf ⍵                 ⍝ not a derv or train: done.
              isop 1⊃⍵:{                          ⍝ derived fn:
                  2=⍴⍵:mop ⍵                      ⍝ monadic operator
                  3=⍴⍵:dop ⍵                      ⍝ dyadic operator.
              }∇¨⍵                                ⍝ formatted subtrees.
              leaf ⍵                              ⍝ neither: give up.
          }
     
          mop←{1 ⍵}∘{                             ⍝ operator with one operand.
              (l m)(land oper)←↓⍉↑⍵               ⍝ derived function components.
              land,oper
          }
     
          dop←{1 ⍵}∘{                             ⍝ operator with two operands.
              (l m r)(land oper rand)←↓⍉↑⍵        ⍝ derived function components.
              0=r:land,oper,rand
              land,oper,'(',rand,')'
          }
     
          leaf←{0 ⍵}∘{                            ⍝ format leaf.
              (⊂⍵)∊pfns,pops:⍵                    ⍝ primitive fn/op.
              (⊂⍵)∊,¨pfns,pops:⍵                  ⍝
              '{'≡⊃⍵:,⍵                           ⍝ unnamed dfn
              1∊'←{'⍷⍵:(1+(,⍵)⍳'←')↓,⍵            ⍝ dfn←{...} → {...}
              '⎕'∊⊃⍵:' ',⍵,' '                    ⍝ system fn/op
              1 repObj ⍵
          }
     
          isfn←{                                  ⍝ is function?
              0=≡⍵:⍵∊pfns                         ⍝ primitive function:
              dfnop ⍵:1                           ⍝ dfn:
              ~(,¨2 3)∨.≡⊂⍴⍵:0                    ⍝ not a derv or train.
              isop 1⊃⍵:1                          ⍝ derv.
              ∇⊃⌽⍵                                ⍝ derv or train.
          }
     
          isop←{                                  ⍝ is operator?
              (⊂⍵)∊pops:1                         ⍝ primitive operator:
              (⊂⍵)∊'⎕S' '⎕R' '⎕OPT':1             ⍝ system op.
              ~dfnop ⍵:0                          ⍝ not a d-op:
              1∊↑'⍺⍺' '⍵⍵'⍷¨⊂⍵                    ⍝ is a d-op.
          }
     
          dfnop←{'}'≡⊃⌽~∘' ',⍵}                   ⍝ dfn or dop
     
          pops←'/\⌿⍀.¨∘⍨&⍣[⌶',⎕UCS(80=⎕DR'')/9056 ⍝ primitive ops.
          pf0←'+-×÷⌊⌈|*⍟<≤=≥>≠∨∧⍱⍲!?~○'           ⍝ primitive fns (scalar).
          pf1←'⊢⊣⌷/⌿\⍀∊⍴↑↓⍳⊂⊃∩∪⊥⊤,⍒⍋⍉⌽⊖⌹⍕⍎⍪≡≢⍷'   ⍝ primitive fns (other).
          pfns←pf0,pf1                            ⍝ primitive fns.
     
          err←⎕SIGNAL∘11                          ⍝ too hard: give up.
          ⍺←⊃⎕RSI
          ⊃⌽trav ⍺.⎕CR ⍵                          ⍝ source of named function ⍵.
      }


      showCol←{ ⍝ Show table in column format
⍝ Each row represents a word
          max←⌈/w←⍬∘⍴∘⍴¨words←trimEnds¨↓⍣(326≠⎕DR ⍵)+⍵
          0∊⍴words:0 0⍴''
          ⎕ML←⎕IO←1
⍝ We can specify the width and minimum space between each word
          ⍺←⍬ ⋄ spacing width←{⍵[⍋⍵]}⍺,(⍳max∧.<⍺),(max∨.<⍺)↓⎕PW
          max w+←spacing
          nw←npl×nrw←⌈(⍴w)÷npl←⌊(width+spacing)÷max
          words←⍉(npl,nrw)⍴nw↑max↑¨words
          (0,-spacing)↓↑,/words
      }

    ∇ r←{options}showRow words;text;sh;blk;sp;pw;⎕ML;⎕IO;n    ⍝ fit a list ⎕PW wide
      ⍎(0=⎕NC'options')/'options←⍬' ⋄ ⎕ML←⎕IO←1
     ⍝ options are Printing Width, Granularity, min Spacing between words
      (pw blk sp)←options,(⍴,options)↓⎕PW,4 1 ⋄ pw←pw-sp
      sh←⊃,/⍴¨,¨text←{~∘' '¨↓⍵}⍣(326≠⎕DR words)+words ⍝ accept matrix of words
      sh←⊃,/⍴¨text←(blk×⌈blk÷⍨sh+sp)↑¨text ⍝ adjust each word's size
      r←0⍴⊂''
      :While 0<⍴text
          r←r,,/(n←1⌈+/pw≥+\sh)↑text ⋄ (text sh)←n↓¨text sh
      :EndWhile
      r←↑r
    ∇

    toMatrix←{⍺←3↓4↑⎕av ⋄ ⎕ml←0 ⋄ (2↑0∊⍴⍵)↓↑1↓¨(s∊⍺)⊂s←(1↑⍺),⍵} ⍝ multiple cut chars allowed

    toVector←{⍺←3↓4↑⎕av ⋄ b←,1,⌽∨\' '≠⌽⍵ ⋄ 1↓b/,⍺,⍵}

    trimEnds←{((∨\b)∧⌽∨\⌽b←' '≠⍵)/⍵}

      repObj←{             ⍝ String representation of object V0.32
⍝ This version accounts for ⎕TC type characters
          ⍵≡⍬:'⍬' ⋄ ⍵≡'':'''''' ⋄ ⍵≡⎕NULL:'⎕NULL'
          scal←0∊rank←+/⍴s←⍴⍵ ⋄ char←>/(simple num)←∧\~(10|⎕DR 1/⍵)∊¨6(0 2)
          n←÷scal⍲1=≡R←⍵   ⍝ no funny objects like ⎕ORs
         ⍝ Refs should normally not be displayed. To allow remove ÷ n the next line
          9=⎕NC'R':÷⍕⍵     ⍝ display refs as they are
     
⍝ Reduce object to 1 item if all same elements
          mod←(0<rank)∧(n=0)∨(5×char)<n←×/s
          mod←mod∧as←char{0∊⍴⍵:1 ⋄ ⍵∧.≡1↑⍵}obj←,⍵ ⍝ as: all the same
          obj←1(↑⍣mod)obj  ⍝ take only 1st? (grab prototype if empty)
          shape←mod{⍵≡,1:',' ⋄ (⍺∨1<⍴⍵)/'⍴',⍨⍕⍵}s
          shape←shape,(encl←simple<as)⍴'⊂'
          ⍺←0              ⍝ recursive call?
     
⍝ Simple scalars and char vector≠⍴1 do not need parens
          parens←⍺∧simple≤(0<⍴shape)∨(rank=1)∧num∨∨/(4↑⎕AV)∊⍵
          (lp rp)←parens⍴¨'()'
          ~simple:rp,⍨lp,shape,encl{⍺⍲'('=1↑⍵:⍵ ⋄ 1↓¯1↓⍵}1↓⊃,/' ',¨{'⎕N'>.=2↑⍵:'(',⍵,')' ⋄ ⍵}¨1 ∇¨obj
     
⍝ Simple objects (char should account for ⎕TC chars et al.)
          ⎕PP←17 ⍝ for numbers
     
          cmpv←{                             ⍝ compress numeric vector
              ∨/e←(0∊s),⍬≡s←⍴v←⍵:⍕e/'⍬',1↑v  ⍝ empty or scalar
              ⍺←4 ⋄ ⍺≥s:⍕v↑⍨⍺⌊s ⋄ ⎕CT←0 ⋄ n←⍴d←1,t←2≠/∆←¯2-/0,v
              t←n↑~∨/t ⋄ r←⍕v↑⍨+/~∨\lim←t∨(⍺↑1 1)⍷d ⍝ parte al frente
              0=n←-+/lim:r
              ⎕ML←1 ⍝ we know we have at least ONE section to deal with
              r,(0∊⍴r)↓⊃,/(n↑1){d v ∆←⍵ ⋄ tail←(t←~0∊b)↓',',⍕v/⍨~b←3>+\d
                  N←⍕+/b ⋄ lp rp←(t⍲⍺)/¨'()'
                  0=d←1↑1↓∆:',',lp,(N,'/',⍕1↑v),rp,tail
                  tail,⍨',',lp,(⍕0~⍨1↑v),'-',((1≠d)/'×',⍨⍕d),'⎕io-⍳',N,rp
              }¨↓⍉↑lim∘⊂¨d v ∆
          }
     
          obj←shape,num ⍺{1↑⍺:cmpv ⍵ ⋄ ⎕ML←1 ⋄ ⎕IO←0 ⋄ QU←{Q,((1+t=Q)/t←⍵),Q←''''}
          ⍝ We have to assume not all characters are available. Those should be:
              Always←⎕A,⎕D,'abcdefghijklmnopqrstuvwxyz_.,:;%!"/=\-+''#$£¢^¿¡(){}[]§@`∣¶'
              Always,←'ÁÂÃÇÈÊËÌÍÎÏÐÒÓÔÕÙÚÛÝþãìðòõÀÄÅÆÉÑÖØÜßàáâäåæçèéêëíîïñùúûüóôöø'
              Always,←'≤≥⌿⍀<>≠∨∧÷×?∊⍴~↑↓⍳○*⌈⌷¨⌊∇∆⍙⍨∘⊂⊃∩∪⊥⊤⌶|⍺⍵¯⍬⍱⍲⍒⍋⍉⌽⊖⍟⌹⍕⍎⍫⍪≡≢⍷⋄←→⍝⎕⍞⍣ '
              ∧/t←⍵∊Always:QU ⍵ ⍝ no special chars?
            ⍝ If only a few chars transform the whole string into ⎕AV
              UCS←{1⌽')(⎕ucs ',⍕cmpv ⎕UCS ⍵}
            ⍝ More than a few; create a mixture of ⎕UCS and 'quotes'
              minsize←3 ⍝ how much special chars to include between sections; that number is subjective
              c∨←minsize>∊⍴¨c←ucs⊂⍨c←1,1↓ucs≠¯1⌽ucs←~t ⍝ consider small groups of ASCII as UCS
             ⍝ ∧/sc←∊c:ucs ⍵ ⍝ are the pieces small enough to be all in ⎕UCS?
              (lp rp)←'()'/⍨¨(1↓⍺)∧1<+/c←1,1↓ucs≠¯1⌽ucs←∊c
              rp,⍨lp,∊{⍺,',',⍵}/(c/ucs){⍺:UCS ⍵ ⋄ QU ⍵}¨c⊂⍵
          }obj
          lp,obj,rp
      }

    ∇ string←toXML value;tag;s;v;⎕PP;repChar;⎕ML;dr
⍝ Turns an APL object into an XML representation
⍝ Simple scalars are turned into CHAR, NUMBER or NULL
⍝ Vectors or rank>1 are turned into ARRAYs of numeric LIST or char STRING
⍝ Characters are translated to be acceptable by ⎕XML
      ⎕PP←17 ⋄ ⎕ML←1
      tag←{'<',s,⍵,'</',s←⍺,'>'}
     ⍝ Some characters cannot go thru ⎕XML, this is normal.
      repChar←∊∘({('<>&'⍳⍵)⊃'&lt;' '&gt;' '&amp;'⍵}¨)
      :If 0∊⍴⍴value ⍝ scalar
          :If 0∊1↑0⍴value ⋄ string←'number'tag⍕value
          :ElseIf ' '∊1↑0⍴value ⋄ string←'char'tag repChar value
          :ElseIf ⎕NULL≡value ⋄ string←'<null/>'
          :ElseIf 1≥|≡value ⋄ 'no Refs allowed'⎕SIGNAL 11
          :Else ⍝ must be enclosed
              string←'enclosed'tag toXML↑value
          :EndIf
      :Else ⍝ not a scalar
          s←'shape'tag⍕⍴value
          :If 2|dr←10|⎕DR value ⋄ string←'list'tag⍕,value
          :ElseIf dr∊0 2 ⋄ string←'string'tag repChar,value
          :Else
              string←∊toXML¨{0∊⍴⍵:⍬⍴⍵ ⋄ ⍵}value
          :EndIf
          string←'array'tag s,string
      :EndIf
    ∇

    ∇ value←fromXML string;mat;shape;⎕ML;type;array;enc;E;n
⍝ Turns an XML string into an APL object
⍝ The string must have been produced by the <toXML> fn above.
      ⎕ML←1 ⋄ E←'enclosed' ⍝ for ⊂
     ⍝ Character strings are transformed into a more suitable format
      :Select type←2⊃,mat←⎕XML⍣(0 2∊⍨10|⎕DR string)+string
     
      :CaseList 'array'E
      ⍝ The array could be enclosed several times
          n←2+enc←+/∧\mat[;2]∊⊂E
          shape←2⊃⎕VFI 3⊃mat[n;]
          value←fromXML¨(mat[n;1]=n↓mat[;1])⊂[1]n 0↓mat
          value←⊂⍣enc+shape⍴↑⍣(1∊⍴value)+value
      :CaseList 'list' 'number'
          value←⍬∘⍴⍣(type≡'number')+2⊃⎕VFI 3⊃,mat
      :CaseList 'string' 'char'
          value←⍬∘⍴⍣(type≡'char')+3⊃,mat
      :Case 'null'
          value←⎕NULL
      :EndSelect
    ∇

    ∇ r←la txtreplace string;⎕IO;N;from;to;Fl;Tl;Fp;i;b;ip;n
⍝ Text Pattern Replace
⍝ la is a 2 elements enclosed strings: what to look for, the replacement string.
⍝ Ex: 'abc' 'xxyyzz' will turn all non overlapping 'abc' substrings into 'xxyyzz'
      (Fl Tl)←⍴¨(from to)←,¨la ⋄ Fp←⍴⎕IO←0 ⋄ N←⍴b←from⍷string←,string
      ⎕SIGNAL Fl↓11
      :If 1∊Fl ⋄ Fp←{⍵/⍳⍴⍵}from=string ⍝ simple case
      :Else
          :While N>i←b⍳1  ⍝ remove overlapping matches
              (Fl↑(b⍳1)↓b)←0 ⋄ Fp,←i
          :EndWhile
      :EndIf
     
⍝ Do it
      :If 0<ip←⍴Fp
          :If 0>n←Fl-Tl   ⍝ do we need to insert spaces?
              b[]←1 ⋄ b[Fp]←1-n ⋄ string←b/string
              Fp←Fp-n×⍳ip ⍝ adjust positions
          :EndIf
          string[,Fp∘.+⍳Tl]←(ip×Tl)⍴to ⍝ replace
          :If n>0         ⍝ remove excess
              b[(Fp+Fl-1)∘.-⍳n]←~b[]←1
              string←b/string
          :EndIf
      :EndIf
      r←string
    ∇

:EndNameSpace