﻿:Namespace  JSutils

    ⎕IO←1 ⋄ ⎕ML←1

    ∇ r←List;desc;names
      :Access Shared Public
      r←{⎕NS ⍬}¨names←'calendar' 'wspeek' 'dinput'
      desc←{r[names⍳⊂⍺].Desc←⍵}  ⍝ short description.
      r.Name←names
      r.Group←⊂'Tools'
      'calendar'desc'Similar to Unix cal'
      'wspeek'desc'Execute expression in temp copy of WS'
      'dinput'desc'Execute/define multi-line D-expression'
      r.Parse←⊂''     ⍝ All fns take raw input buffer as arg.
    ∇

    ∇ r←Run(Cmd Args)
      :Access Shared Public
      :Select Cmd
      :Case 'calendar' ⋄ r←calendar Args
      :Case 'wspeek' ⋄ r←wspeek Args
      :Case 'dinput' ⋄ r←##.THIS dinput Args
      :EndSelect
    ∇

      calendar←{                                      ⍝ Calendar for month or year.
     
          usage←'Usage: ]calendar [month] [year]'     ⍝ usage
     
          month←{                                     ⍝ calendar for month.
              yyyy mm←⍵                               ⍝ year and month number.
              day←days yyyy mm 1                      ⍝ epoch day for 1st of month.
              mms dds←2↑1↓↓⍉date day+¯1+⍳31           ⍝ 31 month and day numbers.
              fmts←2 0∘⍕¨(mm=mms)/dds                 ⍝ char-formatted day numbers.
              dow←7|day                               ⍝ day of week of 1st of month.
              pad←dow↑0↑fmts                          ⍝ start of month padding.
              dmat←↑{⍺,' ',⍵}/dys⍪6 7⍴42↑pad,fmts     ⍝ day matrix.
              head←(mm⊃months),⍺/' ',yfmt yyyy        ⍝ month [year] header.
              ↑(⊂cntr ¯20↑head),↓{(∨/⍵≠' ')⌿⍵}dmat    ⍝ calendar for month ⍵.
          }
     
          days←{                                      ⍝ days since 1899-12-31 (Meeus).
              ⍺←17520902                              ⍝ start of Gregorian calendar.
              yy mm dd h m s ms←7↑⊂[⍳¯1+⍴⍴⍵]⍵         ⍝ ⎕ts-style 7-item date-time.
              D←dd+(0 60 60 1000⊥↑h m s ms)÷86400000  ⍝ day with fractional part.
              Y M←yy mm+¯1 12×⊂mm≤2                   ⍝ Jan, Feb → month 13 14.
              A←⌊Y÷100                                ⍝ century number.
              B←(⍺<0 100 100⊥↑yy mm dd)×(2-A)+⌊A÷4    ⍝ Gregorian calendar correction.
              ¯2416544+D+B+⊃+/⌊365.25 30.6×Y M+4716 1 ⍝ (fractional) days.
          }
     
          date←{⎕ML←0                                 ⍝ ⎕TS from day number (Meeus).
              ⍺←¯53799                                ⍝ UK Gregorian calendar starts.
              qr←{⊂[1+⍳⍴⍴⍵](0,⍺)⊤⍵}                   ⍝ quotient and remainder.
              Z F←1 qr ⍵+2415020                      ⍝
              a←⌊(Z-1867216.25)÷36524.25              ⍝
              A←Z+(Z≥⍺+2415021)×1+a-⌊a÷4              ⍝
              B←A+1524                                ⍝
              C←⌊(B-122.1)÷365.25                     ⍝
              D←⌊C×365.25                             ⍝
              E←⌊(B-D)÷30.6001                        ⍝
              dd df←1 qr(B-D)+F-⌊30.6001×E            ⍝
              mm←E-1+12×E≥14                          ⍝
              yyyy←C-4715+mm>2                        ⍝
              part←60 60 1000 qr⌊0.5+df×86400000      ⍝
              ↑[⎕IO-0.5]yyyy mm dd,part               ⍝
          }
     
          yyyy_mm←{                                   ⍝ month and year.
              2=⍴⍵:(digs¨⍵)infer num¨⍵                ⍝ explicit year and month.
              0=⍴⍵:2↑⎕TS                              ⍝ this year, this month.
              1≠⍴⍵:0                                  ⍝ bad arg vector.
              digs⊃⍵:(num⊃⍵),⊂⍬                       ⍝ year: all months.
              (⊃⎕TS),num⊃⍵                            ⍝ month: month in this year.
          }
     
          infer←{                                     ⍝ guess at month/year combo.
              1=+/⍺:(⍺⍳0)⌽⍵                           ⍝ 2000 feb → 2000 2
              0=+/⍺:0                                 ⍝ june july: bad.
              mm←⍵≤12                                 ⍝ 1..12 interpreted as year.
              1=+/mm:(mm⍳1)⌽⍵                         ⍝ yyyy mm.
              0=+/mm:0                                ⍝ 1914 1918: bad
              2000 0+⌽⍵                               ⍝ 3 4 → 2004 3
          }
     
          y2k←{                                       ⍝ 50-year window around yy.
              ⍵≥100:⍵                                 ⍝ full year yyyy?
              now←100|⊃⎕TS                            ⍝ this year.
              ⍵+1900+100×50>|now-⍵                    ⍝ 98→1998  12→1012
          }
     
          num←{                                       ⍝ number from char vector.
              digs ⍵:⍎⍵                               ⍝ '1984' → 1984
              mms←(⍴⍵)↑¨mons                          ⍝ jan feb ...
              mm←mms⍳⊂lcase ⍵                         ⍝ feb → 2
              mm×mm≤12                                ⍝ fev: bad month.
          }
     
          cntr←{(⌈0.5×+/∧\' '=⍵)⌽⍵}                   ⍝ centred text.
          join←{⍉↑(↓⍉⍺),'   ',↓⍉⍵}                    ⍝ month joiner.
          digs←{∧/⍵∊⎕D}                               ⍝ digit vector.
          yfmt←{(-4⌈⍴⍕⍵)↑'000',⍕⍵}                    ⍝ leading 0s for years <1000.
     
          dys←'Su' 'Mo' 'Tu' 'We' 'Th' 'Fr' 'Sa'      ⍝ day-of-week column headers.
          Q0←'January' 'February' 'March    '~¨' '    ⍝ 1st quarter month names.
          Q1←'April  ' 'May     ' 'June     '~¨' '    ⍝ 2nd   ..      ..    ..
          Q2←'July   ' 'August  ' 'September'~¨' '    ⍝ 3rd   ..      ..    ..
          Q3←'October' 'November' 'December '~¨' '    ⍝ 4th   ..      ..    ..
          months←Q0,Q1,Q2,Q3                          ⍝ month names.
          mons←lcase¨months                           ⍝ for case-free comparison.
     
          wds←list rmcm ⍵                             ⍝ blank-separated words.
          2<⍴wds:usage                                ⍝ too many args: give up.
     
          yyyy mm←⌽y2k\⌽yyyy_mm wds                   ⍝ year and month number.
          0∊yyyy mm:usage                             ⍝ bad date.
          ~mm≡⍬:1 month yyyy mm                       ⍝ calendar for single month.
          year←4 3⍴{0 month yyyy ⍵}¨⍳12               ⍝ each month for given year.
          head←cntr ¯66↑yfmt yyyy                     ⍝ centred year header.
          head⍪,[⍳2]↑join/year                        ⍝ calendar for year ⍵.
      }

      wspeek←{⎕IO←0                               ⍝ Execute expression in temp copy of WS
     
          usage←'Usage: ]WSpeek wsid [expr ...]'  ⍝ usage.
     
          wsid tail←↑{                            ⍝ wsid and remainder.
              ((⍺≡,'.')⊃⍺ ⎕WSID)⍵                 ⍝ '.' means ⎕wsid.
          }/{                                     ⍝ first word and remainder.
              raw←rmqt ⍵                          ⍝ ignoring quoted '"' chars.
              ~'"'∊raw:vect\head list rmcm ⍵      ⍝ wsid is first word.
              1↓¨('"'=⊃↓↑raw ⍵)⊂⍵                 ⍝ "-delimited file name.
          }rmcm ⍵
     
          wsid∧.=' ':usage                        ⍝ ]ws
          expr←(tail∧.=' ')⊃tail'⍎⎕lx'           ⍝ ]ws wsid → ]ws wsid ⍎⎕lx
                                                ⍝ ]ws wsid expr ...
          ⎕LX ⎕WSID←⊂''                           ⍝ localise session vars.
          nsv1←'⎕ct' '⎕io' '⎕div' '⎕ml'           ⍝ namespace extent ...
          nsv2←'⎕pp' '⎕rl' '⎕rtl' '⎕wx'           ⍝ ... system variables.
          nsv3←'⎕lx' '⎕wsid' '⎕trap'              ⍝ ws extent vars.
     
          copy←'⎕cy''',wsid,''' ⋄ '               ⍝ copy command.
          lasv←'(##.(↑nsv1,nsv2,nsv3))'           ⍝ left argt for sysvar copy.
          11::'Can''t copy "',wsid,'"'            ⍝ copy failed.
          ref←(⎕NS'')⍎copy,lasv,copy,'⎕this'      ⍝ ref to copied ws.
          0::(⊃⎕DM)⎕SIGNAL ⎕EN                    ⍝ pass back all errors.
          6::0 0⍴0                                ⍝ ignore value error.
          _←ref.⎕DF']WSpeek ',wsid,' #'           ⍝ display form, for ]ws wsid ⎕this
          ~expr≡,'→':{⍵}ref⍎expr                  ⍝ single expr: evaluate in copy of ws.
          trim←{('·'≠⍵){⍺\⍺/⍵}⍵}                  ⍝ replace white dots with blanks.
          {                                       ⍝ ⍎⍞ loop until →.
              0::∇ ⎕←lcase(⊃⎕DM)~'⍎'              ⍝ error:: report and continue.
              expr←trim{⍞}⍞←4/'·'                 ⍝ next expr
              ' '∧.=trim rmcm expr:∇ ⍵            ⍝ ignore null expr.
              rslt←ref⍎expr                       ⍝ evaluate expr in copy of ws.
              '←'∊rmcm rmqt expr:∇ ⍵              ⍝ assignment: continue.
              ∇ ⎕←rslt                            ⍝ display rslt and continue.
          }⍬
      }

      dinput←{⎕IO←0                               ⍝ Multi-line D-expression/defn.
     
          accum←{                                 ⍝ accumulation loop
              toks←rmqt rmcm⊃⌽⍺                   ⍝ line tokens.
              (,'→')≡toks~' ':(,⊂'')'0 0⍴0'       ⍝ →: abort.
              null←'{'=⊃⌽'{',toks~' '             ⍝ null line.
              null<0=depth ⍵:⍺ ⍵                  ⍝ depth=0: done.
              defn←'←'∊toks                       ⍝ guarded line.
              gard←':'∊toks                       ⍝ local defn.
              more←null∨defn∨gard                 ⍝ result: exdent next line.
              dent←(depth ⍵)-~more                ⍝ indentation.
              next←blnks ask dent                 ⍝ next line.
              (⍺,⊂next)∇ ⍵,' ⋄ ',rmcm next        ⍝ extended input.
          }
     
          ask←{{⍞}⍞←(4×⍵+1)/'·'}                  ⍝ prompt with white dots.
          depth←{0⌈-/+/'{}'∘.=rmqt ⍵}             ⍝ {-expression depth.
          blnks←{(⍵≠'·'){⍺\⍺/⍵}⍵}                 ⍝ subs white dots with blanks.
          align←{(4⌊+/∧\' '=⍵)↓⍵}                 ⍝ remove indentation from defn.
     
          line0←{' '∧.=⍵:blnks ask 0 ⋄ ⍵}⍵        ⍝ prompt for 1st line if not given.
          defn expr←(,⊂line0)accum rmcm line0     ⍝ accumulate expr/defn lines.
          lines←{(∨\~(rmcm¨⍵~¨' ')∊⊂'')/⍵}defn    ⍝ excluding leading comment lines.
          0::⎕SIGNAL ⎕EN                          ⍝ pass back all errors.
          ≥/(⊃lines)⍳'←{':⍺⍎expr                  ⍝ expression: exec in calling space.
          rslt←⍺.⎕FX align¨lines                  ⍝ fix dfn defn in calling space.
          ⍎(0=⊃0⍴rslt)⊃'0 0⍴0' '⎕signal 2'        ⍝ syntax error if failure to fix.
      }

    ∇ r←lev Help Cmd
      :Access Shared Public
      :Select Cmd
      :Case 'calendar'
          r←,⊂'Usage: ]calendar [month] [year]'
          r,←⊂''
          r,←⊂'For example:'
          r,←⊂''
          r,←⊂'      ]cal            ⍝ this month.'
          r,←⊂'      ]cal April      ⍝ specific month this year.'
          r,←⊂'      ]cal sept 1752  ⍝ specific month/year.'
          r,←⊂'      ]cal 2012       ⍝ calendar for whole year.'
          r,←⊂'      ]cal dec 78     ⍝ Y2K-style 50-year window.'
          r,←⊂'      ]cal 1 76       ⍝ numeric month.'
          r,←⊂'      ]cal 3 4        ⍝ intrepreted as March 2004'
          r,←⊂''
          r←↑r
      :Case 'wspeek'
          r←,⊂'Usage: ]WSpeek wsid [expr ...]       ⍝ expr defaults to ⍎⎕LX'
          r,←⊂''
          r,←⊂'Executes expression in temp copy of WS. As its name suggests,'
          r,←⊂'WSpeek is used to view,  rather than to change, a saved work-'
          r,←⊂'space;  any changes made in the copy are discarded on termin-'
          r,←⊂'ation of the command.    Wsid ''.'' means the saved copy of the'
          r,←⊂'current workspace.'
          :If lev>0
              r,←⊂''
              r,←⊂'For example:'
              r,←⊂''
              r,←⊂'      ]wsp arachnid                 ⍝ NB: use "Alt+g a" to quit'
              r,←⊂'      ]wsp dfns  0 disp queens 5    ⍝ exec expr in copy of dfns'
              r,←⊂'      ]wsp dfns  tree ⎕this         ⍝ ⎕this is root space of copy'
              r,←⊂'      ]wsp tube  paris.⎕ed''source''  ⍝ _view_ var in copy of ws'
              r,←⊂'      ]wsp dfns  cols ⎕nl 3 4       ⍝ fns and ops in dfns.dws'
              r,←⊂'      ]wsp . myfn                   ⍝ saved version of myfn'
              r,←⊂''
              r,←⊂'NB:  The ⎕LX in some workspaces may end by calling ⎕OFF, thus'
              r,←⊂'terminating the APL session.'
              r,←⊂''
              r,←⊂'Special expression ''→'' starts a simple interactive ⍎-⍞ loop,'
              r,←⊂'which terminates on input of a single ''→'' character:'
              r,←⊂''
              r,←⊂'      ]wsp dfns →            ⍝ interactive loop in copy.'
              r,←⊂'····'
              r,←⊂'····fns ops ← ⎕nl¨ -3 4      ⍝ vars for duration of loop.'
              r,←⊂'····'
              r,←⊂'····⍴fns                     ⍝ simple expression in copy.'
              r,←⊂'178'
              r,←⊂'····lines ← ⊃∘⍴∘⎕cr          ⍝ temporary function definition.'
              r,←⊂'····'
              r,←⊂'····+/lines¨fns,ops          ⍝ using all of the above.'
              r,←⊂'6014'
              r,←⊂'····→                        ⍝ termination of loop'
              r,←⊂''
              r,←⊂'The "white dots" ··· in the prompt are a reminder that the ⍎-⍞'
              r,←⊂'loop is active and that you should type → to return to regular'
              r,←⊂'session input (losing any changes in the copy).'
          :EndIf
          r,←⊂''
          r←↑r
      :Case 'dinput'
          r←,⊂'Execute or define multi-line D-expression'
          r,←⊂''
          r,←⊂'For example:'
          r,←⊂''
          r,←⊂'      ]Dinput    ⍝ multi-line expression'
          r,←⊂'····{            ⍝ dup:'
          r,←⊂'········⍵ ⍵'
          r,←⊂'····}{           ⍝ twice:'
          r,←⊂'········⍺⍺ ⍺⍺ ⍵'
          r,←⊂'····}7'
          r,←⊂' 7 7  7 7 '
          r,←⊂''
          :If lev>0
              r,←⊂'      ]din       ⍝ multi-line definition'
              r,←⊂'····natsum←{         ⍝ nats ⍺+⍵'
              r,←⊂'········⍺=0:⍵        ⍝ 0+⍵ → ⍵'
              r,←⊂'········(⍺-1)∇ 1+⍵   ⍝ (⍺+1)+⍵ → ⍺+(1+⍵)'
              r,←⊂'····}'
              r,←⊂''
              r,←⊂'The "white dots" ····, which will be replaced with blanks, are a'
              r,←⊂'reminder that you are in an expression accumulation loop.'
              r,←⊂''
              r,←⊂'To abort the loop at any time, enter a single "→" character.'
              r,←⊂''
              r,←⊂'When entering a non-trivial expression, it may be easier to type'
              r,←⊂'the ]din command and then press Enter several times to produce a'
              r,←⊂'sufficient block of input prompts:'
              r,←⊂''
              r,←⊂'      ]din'
              r,←⊂'····         ┐ Then type over these lines in any order'
              r,←⊂'····         │ before pressing <Enter> to evaluate the'
              r,←⊂'····         ┘ expression.'
              r,←⊂''
              r,←⊂'You can amend a previously submitted  multi-line input simply by'
              r,←⊂'making corrections in the session history and then selecting the'
              r,←⊂'complete expression, including its initial "]din", before press-'
              r,←⊂'ing <Enter> to resend it. Experiment with the above examples.'
              r,←⊂''
          :EndIf
          r←↑r
      :EndSelect
    ∇

    COM←'⍝'                                 ⍝ prob with embedded '⍝' in dfn.
    lc←'abcdefghijklmnopqrstuvwxyz'         ⍝ lower case alphabet
    uc←'ABCDEFGHIJKLMNOPQRSTUVWXYZ'         ⍝ upper case alphabet
    lcase←{(lc,⎕AV)[(uc,⎕AV)⍳⍵]}            ⍝ lower-casification,
    rmcm←{(∧\~(COM=⍵)∧qmask ⍵)/⍵}           ⍝ ignoring unquoted comment.
    qmask←{{⍵∨¯1⌽⍵}~≠\⍵=''''}               ⍝ mask of unquoted chars.
    head←{(⊃⍵)(1↓⍵)}                        ⍝ head and tail of list.
    rmqt←{(qmask ⍵){⍺\⍺/⍵}⍵}                ⍝ with blanks for quoted chars.
    list←{(~∘' '¨(1,' '=⍵)⊂' ',⍵)~⊂''}      ⍝ blank-separated word list.
    vect←{¯2↓↑{⍺,' ',⍵}/⍵,'' ''}            ⍝ enlist of blank-separated words.

:EndNamespace